/*
  postgresql.c -- Emacs Lisp binding to libpq.so
  Copyright (C) 2000 Electrotechnical Laboratory, JAPAN.
  Licensed to the Free Software Foundation.
*/

#include <config.h>
#include "lisp.h"
#include "sysdep.h"

#include <libpq-fe.h>
#include "postgresql.h"

/* #define HAVE_POSTGRESQLV7 /* Until configure gets fixed */

static Lisp_Object VXPGHOST;
static Lisp_Object VXPGUSER;
static Lisp_Object VXPGOPTIONS;
static Lisp_Object VXPGPORT;
/* static Lisp_Object VXPGTTY; /* This needs to be blanked! */
static Lisp_Object VXPGDATABASE;
static Lisp_Object VXPGREALM;
#ifdef MULE
static Lisp_Object VXPGCLIENTENCODING;
#endif /* MULE */

/* Other variables we don't provide an interface for:
   PGAUTHTYPE -- not used after PostgreSQL 6.5
   PGGEQO
   PGCOSTINDEX
   PGCOSTHEAP
   PGTZ
   PGDATESTYLE
*/

static Lisp_Object Qpqdb, Qpquser, Qpqpass, Qpqhost, Qpqport, Qpqtty;
static Lisp_Object Qpqoptions, Qpqstatus, Qpqerrormessage, Qpqbackendpid;

static Lisp_Object Qpgres_empty_query, Qpgres_command_ok, Qpgres_tuples_ok;
static Lisp_Object Qpgres_copy_out, Qpgres_copy_in, Qpgres_bad_response;
static Lisp_Object Qpgres_nonfatal_error, Qpgres_fatal_error;

/****/

/* PGconn is an opaque object and we need to be able to store them in
   Lisp code because libpq supports multiple connections.
*/
Lisp_Object Qpgconnp;

static Lisp_Object
make_pgconn (struct Lisp_PGconn *pgconn)
{
  Lisp_Object lisp_pgconn;
  XSETPGCONN (lisp_pgconn, pgconn);
  return lisp_pgconn;
}

static Lisp_Object
mark_pgconn (Lisp_Object obj, void (*markobj) (Lisp_Object))
{
  return Qnil;
}

static void
print_pgconn (Lisp_Object obj, Lisp_Object printcharfun, int escapeflag)
{
  char buf[256];
  PGconn *pgc;
  ConnStatusType cst;
  char *host="", *db="", *user="", *port="";

  pgc = (XPGCONN (obj))->pgconn;
  
  if ((cst = PQstatus (pgc)) == CONNECTION_OK)
    {
      if (!(host = PQhost(pgc)))
	host = "";
      port = PQport(pgc);
      db = PQdb(pgc);
      if (!(user = PQuser(pgc)))
	user = "";
      sprintf(buf, "#<PGconn %s:%s %s/%s>", /* evil! */
	      !strlen(host) ? "localhost" : host,
	      port,
	      user,
	      db);
    }
  else
    strcpy(buf, "#<PGconn BAD>");

  if (print_readably)
    error ("printing unreadable object %s", buf);
  else
    write_c_string (buf, printcharfun);
}

static struct Lisp_PGconn *
allocate_pgconn (void)
{
  struct Lisp_PGconn *pgconn = alloc_lcrecord_type (struct Lisp_PGconn,
						    lrecord_pgconn);
  pgconn->pgconn = (PGconn *)NULL;
  return pgconn;
}

static void
finalize_pgconn (void *header, int for_disksave)
{
  struct Lisp_PGconn *pgconn = (struct Lisp_PGconn *) header;

  if (for_disksave)
    signal_simple_error ("Can't dump an emacs containing PGconn objects",
                         make_pgconn (pgconn));

  if (pgconn->live)
    PQfinish (pgconn->pgconn);
}

DEFINE_LRECORD_IMPLEMENTATION ("pgconn", pgconn,
			       mark_pgconn, print_pgconn, finalize_pgconn,
			       NULL, NULL, struct Lisp_PGconn);

/****/

/* PGresult is an opaque object and we need to be able to store them in
   Lisp code.
*/
Lisp_Object Qpgresultp;

static Lisp_Object
make_pgresult (struct Lisp_PGresult *pgresult)
{
  Lisp_Object lisp_pgresult;
  XSETPGRESULT (lisp_pgresult, pgresult);
  return lisp_pgresult;
}

static Lisp_Object
mark_pgresult (Lisp_Object obj, void (*markobj) (Lisp_Object))
{
  return Qnil;
}

static void
print_pgresult (Lisp_Object obj, Lisp_Object printcharfun, int escapeflag)
{
  char buf[1024];
  PGresult *res;

  res = (XPGRESULT(obj))->pgresult;

  sprintf (buf, "#<PGresult %s - %s>",  /* evil! */
		 PQresStatus (PQresultStatus (res)),
		 PQcmdStatus (res));

  if (print_readably)
    error ("printing unreadable object %s", buf);
  else
    write_c_string (buf, printcharfun);
}

static struct Lisp_PGresult *
allocate_pgresult (void)
{
  struct Lisp_PGresult *pgresult = alloc_lcrecord_type (struct Lisp_PGresult,
							lrecord_pgresult);
  pgresult->pgresult = (PGresult *)NULL;
  return pgresult;
}

static void
finalize_pgresult (void *header, int for_disksave)
{
  struct Lisp_PGresult *pgresult = (struct Lisp_PGresult *) header;

  if (for_disksave)
    signal_simple_error ("Can't dump an emacs containing PGresult objects",
                         make_pgresult (pgresult));

  if (pgresult->live)
    PQclear (pgresult->pgresult);
}

DEFINE_LRECORD_IMPLEMENTATION ("pgresult", pgresult,
			       mark_pgresult, print_pgresult, finalize_pgresult,
			       NULL, NULL, struct Lisp_PGresult);


/* There are four ways (as of PostgreSQL v7) to connect to a database.
   Two of them, PQsetdb and PQsetdbLogin, are deprecated.  Both of those
   routines take a number of positional parameters and are better done in Lisp.
   Note that PQconnectStart does not exist prior to v7.
*/

/* PQconnectdb Makes a new connection to a backend.
PGconn *PQconnectdb(const char *conninfo)
*/

DEFUN ("pq-connectdb", Fpq_connectdb, 1, 1, 0, /*
Make a new connection to a PostgreSQL backend.
*/
	(conninfo))
{
  PGconn *P;
  struct Lisp_PGconn *lisp_pgconn;

  CHECK_STRING (conninfo);

  P = PQconnectdb ((char *)XSTRING_DATA (conninfo));
  lisp_pgconn = allocate_pgconn();
  lisp_pgconn->pgconn = P;
  lisp_pgconn->live = 1;

  return make_pgconn (lisp_pgconn);
}

/* PQconnectStart Makes a new asynchronous connection to a backend.
PGconn *PQconnectStart(const char *conninfo)
*/

#ifdef HAVE_POSTGRESQLV7
DEFUN ("pq-connect-start", Fpq_connect_start, 1, 1, 0, /*
Make a new asynchronous connection to a PostgreSQL backend.
*/
	(conninfo))
{
  PGconn *P;
  struct Lisp_PGconn *lisp_pgconn;

  CHECK_STRING (conninfo);

  P = PQconnectStart ((char *)XSTRING_DATA (conninfo));

  lisp_pgconn = allocate_pgconn();
  lisp_pgconn->pgconn = P;
  lisp_pgconn->live = 1;

  return make_pgconn (lisp_pgconn);
}
#endif /* HAVE_POSTGRESQLV7 */

/* PQfinish Close the connection to the backend. Also frees memory
       used by the PGconn object.
void PQfinish(PGconn *conn)
*/
DEFUN ("pq-finish", Fpq_finish, 1, 1, 0, /*
Close the connection to the backend.
Note:  This entry point is provided for completeness, but its use is not
recommended.  The preferred way to delete connections is to let the Emacs
garbage collection mechanism do it.
*/
	(conn))
{
  PGconn *pgconn;

  CHECK_PGCONN (conn);
  pgconn = (XPGCONN (conn))->pgconn;

  PQfinish (pgconn);
  /* #### PQfinish deallocates the PGconn structure, so we now have a
     dangling pointer. */
  XSETPGCONN (conn, (PGconn *)NULL); /* will this work? */
  return Qnil;
}

/* PQreset Reset the communication port with the backend.
void PQreset(PGconn *conn)
*/
DEFUN ("pq-reset", Fpq_reset, 1, 1, 0, /*
Reset the connection to the backend.
This function will close the connection to the backend and attempt to
reestablish a new connection to the same postmaster, using all the same
parameters previously used.  This may be useful for error recovery if a
working connection is lost.
*/
	(conn))
{
  PGconn *pgconn;

  CHECK_PGCONN (conn);
  pgconn = (XPGCONN (conn))->pgconn;

  PQreset (pgconn);

  return Qnil;
}

/* accessor function for the PGconn object */
DEFUN ("pq-pgconn", Fpq_pgconn, 2, 2, 0, /*
Accessor function for the PGconn object.
Currently recognized symbols for the field:
pq:DB            Database name
pq:USER          Database user name
pq:PASS          Database user's password
pq:HOST          Hostname of PostgreSQL backend connected to
pq:PORT          TCP port number of connection
pq:TTY           Debugging TTY (not used in Emacs)
pq:OPTIONS       Additional backend options
pq:STATUS        Connection status (either OK or BAD)
pq:ERROR-MESSAGE Last error message from the backend
pq:BACKEND-PID   Process ID of backend process
*/
	(conn, field))
{
  PGconn *pgconn;

  CHECK_PGCONN (conn);
  pgconn = (XPGCONN (conn))->pgconn;

  if (!pgconn)
    error ("Stale connection");
  else if (EQ(field, Qpqdb))
    /* PQdb Returns the database name of the connection.
       char *PQdb(PGconn *conn)
     */
    return build_string (PQdb(pgconn));
  else if (EQ (field, Qpquser))
    /* PQuser Returns the user name of the connection.
       char *PQuser(PGconn *conn)
     */
    return build_string (PQuser(pgconn));
  else if (EQ (field, Qpqpass))
    /* PQpass Returns the password of the connection.
       char *PQpass(PGconn *conn)
     */
    return build_string (PQpass(pgconn));
  else if (EQ (field, Qpqhost))
    /* PQhost Returns the server host name of the connection.
       char *PQhost(PGconn *conn)
     */
    return build_string (PQhost(pgconn));
  else if (EQ (field, Qpqport))
    {
      char *p;
      /* PQport Returns the port of the connection.
         char *PQport(PGconn *conn)
       */
      if ((p = PQport(pgconn)))
	return make_int(atoi(p));
      else
        return make_int(-1);
    }
  else if (EQ (field, Qpqtty))
    /* PQtty Returns the debug tty of the connection.
       char *PQtty(PGconn *conn)
     */
    return build_string (PQtty(pgconn));
  else if (EQ (field, Qpqoptions))
  /* PQoptions Returns the backend options used in the connection.
     char *PQoptions(PGconn *conn)
   */
    return build_string (PQoptions(pgconn));
  else if (EQ (field, Qpqstatus))
    /* PQstatus Returns the status of the connection. The status can be
       CONNECTION_OK or CONNECTION_BAD.
       ConnStatusType PQstatus(PGconn *conn)
     */
    return (PQstatus(pgconn) == CONNECTION_OK ? Qt : Qnil);
  else if (EQ (field, Qpqerrormessage))
    /* PQerrorMessage Returns the error message most recently generated
       by an operation on the connection.
       char *PQerrorMessage(PGconn* conn);
     */
    return build_string (PQerrorMessage(pgconn));
  else if (EQ (field, Qpqbackendpid))
    /* PQbackendPID Returns the process ID of the backend server handling
       this connection.
       int PQbackendPID(PGconn *conn);
     */
    return make_int (PQbackendPID(pgconn));
  else
    error ("bad PGconn accessor");
}

/* Query functions */
DEFUN ("pq-exec", Fpq_exec, 2, 2, 0, /*
Submit a query to Postgres and wait for the result.
*/
	(conn, query))
{
  PGconn *pgconn;
  struct Lisp_PGresult *lisp_pgresult;
  PGresult *pgresult;

  CHECK_PGCONN (conn);
  CHECK_STRING (query);

  pgconn = (XPGCONN (conn))->pgconn;

  pgresult = PQexec (pgconn, XSTRING_DATA (query));
  lisp_pgresult = allocate_pgresult();
  lisp_pgresult->pgresult = pgresult;
  lisp_pgresult->live = 1;

  return make_pgresult (lisp_pgresult);
}

DEFUN ("pq-result-status", Fpq_result_status, 1, 1, 0, /*
Return result status of the query.
*/
	(result))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  pgr = (XPGRESULT (result))->pgresult;

  switch (PQresultStatus (pgr)) {
  case PGRES_EMPTY_QUERY: return Qpgres_empty_query;
  case PGRES_COMMAND_OK: return Qpgres_command_ok;
  case PGRES_TUPLES_OK: return Qpgres_tuples_ok;
  case PGRES_COPY_OUT: return Qpgres_copy_out;
  case PGRES_COPY_IN: return Qpgres_copy_in;
  case PGRES_BAD_RESPONSE: return Qpgres_bad_response;
  case PGRES_NONFATAL_ERROR: return Qpgres_nonfatal_error;
  case PGRES_FATAL_ERROR: return Qpgres_fatal_error;
  default: abort ();
  }
}

DEFUN ("pq-res-status", Fpq_res_status, 1, 1, 0, /*
Return stringified result status of the query.
*/
	(result))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  pgr = (XPGRESULT (result))->pgresult;

  return build_string (PQresStatus(PQresultStatus (pgr)));
}

/* Sundry PGresult accessor functions */
DEFUN ("pq-result-error-message", Fpq_result_error_message, 1, 1, 0, /*
Return last message associated with the query.
*/
	(result))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  pgr = (XPGRESULT (result))->pgresult;

  return build_string (PQresultErrorMessage(pgr));
}

DEFUN ("pq-ntuples", Fpq_ntuples, 1, 1, 0, /*
Return the number of tuples (instances) in the query result.
*/
	(result))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  pgr = (XPGRESULT (result))->pgresult;

  return make_int (PQntuples (pgr));
}

DEFUN ("pq-nfields", Fpq_nfields, 1, 1, 0, /*
Return the number of fields (attributes) in each tuple of the query result.
*/
	(result))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  pgr = (XPGRESULT (result))->pgresult;

  return make_int (PQnfields (pgr));
}

DEFUN ("pq-binary-tuples", Fpq_binary_tuples, 1, 1, 0, /*
Return t if the query result contains binary data, nil otherwise.
*/
	(result))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  pgr = (XPGRESULT (result))->pgresult;

  return (PQbinaryTuples (pgr)) ? Qt : Qnil;
}

DEFUN ("pq-fname", Fpq_fname, 2, 2, 0, /*
Return the field (attribute) name associated with the given field index.
Field indices start at 0.
*/
	(result, field_index))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  CHECK_INT (field_index);
  pgr = (XPGRESULT (result))->pgresult;

  return build_string (PQfname(pgr, XINT (field_index)));
}

DEFUN ("pq-fnumber", Fpq_fnumber, 2, 2, 0, /*
Return the number of fields (attributes) in each tuple of the query result.
*/
	(result, field_name))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  CHECK_STRING (field_name);
  pgr = (XPGRESULT (result))->pgresult;

  return make_int (PQfnumber (pgr, XSTRING_DATA (field_name)));
}

DEFUN ("pq-ftype", Fpq_ftype, 2, 2, 0, /*
Return the field type associated with the given field index.
The integer returned is the internal coding of the type.  Field indices
start at 0.
*/
	(result, field_num))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  CHECK_INT (field_num);
  pgr = (XPGRESULT (result))->pgresult;

  return make_int (PQftype (pgr, XINT (field_num)));
}

DEFUN ("pq-fsize", Fpq_fsize, 2, 2, 0, /*
Return the field size in bytes associated with the given field index.
Field indices start at 0.
*/
	(result, field_index))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  CHECK_INT (field_index);
  pgr = (XPGRESULT (result))->pgresult;

  return make_int (PQftype (pgr, XINT (field_index)));
}

DEFUN ("pq-get-value", Fpq_get_value, 3, 3, 0, /*
Return a single field (attribute) value of one tuple of a PGresult.
Tuple and field indices start at 0.
*/
	(result, tup_num, field_num))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  CHECK_INT (tup_num);
  CHECK_INT (field_num);
  pgr = (XPGRESULT (result))->pgresult;

  return build_string (PQgetvalue (pgr, XINT (tup_num), XINT (field_num)));
}

DEFUN ("pq-cmd-status", Fpq_cmd_status, 1, 1, 0, /*
Returns the command status string from the SQL command that generated the result.
*/
	(result))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  pgr = (XPGRESULT (result))->pgresult;

  return build_string (PQcmdStatus (pgr));
}

DEFUN ("pq-cmd-tuples", Fpq_cmd_tuples, 1, 1, 0, /*
Returns the number of rows affected by the SQL command
*/
	(result))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  pgr = (XPGRESULT (result))->pgresult;

  return build_string (PQcmdTuples (pgr));
}

DEFUN ("pq-oid-status", Fpq_oid_status, 1, 1, 0, /*
Returns a string with the object id of the tuple inserted.
Returns an empty string if the SQL command was not an insertion
*/
	(result))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  pgr = (XPGRESULT (result))->pgresult;

  return build_string (PQoidStatus (pgr));
}


void
syms_of_postgresql(void)
{
  defsymbol (&Qpgconnp, "pgconnp");
  defsymbol (&Qpgresultp, "pgresultp");

  /* Fields of PGconn */
  defsymbol (&Qpqdb, "pq::DB");
  defsymbol (&Qpquser, "pq::USER");
  defsymbol (&Qpqpass, "pq::PASS");
  defsymbol (&Qpqhost, "pq::HOST");
  defsymbol (&Qpqport, "pq::PORT");
  defsymbol (&Qpqtty, "pq::TTY");
  defsymbol (&Qpqoptions, "pq::OPTIONS");
  defsymbol (&Qpqstatus, "pq::STATUS");
  defsymbol (&Qpqerrormessage, "pq::ERROR-MESSAGE");
  defsymbol (&Qpqbackendpid, "pq::BACKEND-PID");

  /* Query status results */
  defsymbol (&Qpgres_empty_query, "pgres::EMPTY-QUERY");
  defsymbol (&Qpgres_command_ok, "pgres::COMMAND-OK");
  defsymbol (&Qpgres_tuples_ok, "pgres::TUPLES-OK");
  defsymbol (&Qpgres_copy_out, "pgres::COPY-OUT");
  defsymbol (&Qpgres_copy_in, "pgres::COPY-IN");
  defsymbol (&Qpgres_bad_response, "pgres::BAD-RESPONSE");
  defsymbol (&Qpgres_nonfatal_error, "pgres::NONFATAL-ERROR");
  defsymbol (&Qpgres_fatal_error, "pgres::FATAL-ERROR");

#ifdef HAVE_POSTGRESQLV7
  DEFSUBR (Fpq_connect_start);
#endif /* HAVE_POSTGRESQLV7 */
  DEFSUBR (Fpq_connectdb);
  DEFSUBR (Fpq_finish);
  DEFSUBR (Fpq_reset);
  DEFSUBR (Fpq_pgconn);

  DEFSUBR (Fpq_exec);
  DEFSUBR (Fpq_result_status);
  DEFSUBR (Fpq_res_status);
  DEFSUBR (Fpq_result_error_message);
  DEFSUBR (Fpq_ntuples);
  DEFSUBR (Fpq_nfields);
  DEFSUBR (Fpq_binary_tuples);
  DEFSUBR (Fpq_fname);
  DEFSUBR (Fpq_fnumber);
  DEFSUBR (Fpq_ftype);
  DEFSUBR (Fpq_fsize);
  /***/
  DEFSUBR (Fpq_get_value);
  DEFSUBR (Fpq_cmd_status);
  DEFSUBR (Fpq_cmd_tuples);
  DEFSUBR (Fpq_oid_status);
}

void
vars_of_postgresql(void)
{
  char *p;

  Fprovide (intern ("postgresql"));

  if ((p = getenv("PGHOST")))
    {
      VXPGHOST = build_string(p);
    }
  else
    {
      VXPGHOST = Qnil;
    }
  DEFVAR_LISP ("pg-host", &VXPGHOST /*
Default PostgreSQL server name.
If not set, the server running on the local host is used.  The
initial value is set from the PGHOST environment variable.
*/ );

  if ((p = getenv("PGUSER")))
    {
      VXPGUSER = build_string(p);
    }
  else
    {
      VXPGUSER = Qnil;
    }
  DEFVAR_LISP ("pg-user", &VXPGUSER /*
Default PostgreSQL user name.
This value is used when connecting to a database for authentication.
The initial value is set from the PGUSER environment variable.
*/ );

  if ((p = getenv("PGOPTIONS")))
    {
      VXPGOPTIONS = build_string(p);
    }
  else
    {
      VXPGOPTIONS = Qnil;
    }
  DEFVAR_LISP ("pg-options", &VXPGOPTIONS /*
Default PostgreSQL user name.
This value is used when connecting to a database for authentication.
The initial value is set from the PGUSER environment variable.
*/ );

  if ((p = getenv("PGPORT")))
    {
      VXPGPORT = make_int(atoi(p));
    }
  else
    {
      VXPGPORT = Qnil;
    }
  DEFVAR_LISP ("pg-port", &VXPGPORT /*
Default port to connect to PostgreSQL backend.
This value is used when connecting to a database.
The initial value is set from the PGPORT environment variable.
*/ );

  if ((p = getenv("PGDATABASE")))
    {
      VXPGDATABASE = build_string(p);
    }
  else
    {
      VXPGDATABASE = Qnil;
    }
  DEFVAR_LISP ("pg-database", &VXPGDATABASE /*
Default database to connect to.
The initial value is set from the PGDATABASE environment variable.
*/ );

  if ((p = getenv("PGREALM")))
    {
      VXPGREALM = build_string(p);
    }
  else
    {
      VXPGREALM = Qnil;
    }
  DEFVAR_LISP ("pg-realm", &VXPGREALM /*
Default kerberos realm to use for authentication.
The initial value is set from the PGREALM environment variable.
*/ );

#ifdef MULE
  if ((p = getenv("PGCLIENTENCODING")))
    {
      VXPGCLIENTENCODING = build_string(p);
    }
  else
    {
      VXPGCLIENTENCODING = Qnil;
    }
  DEFVAR_LISP ("pg-clientencoding", &VXPGCLIENTENCODING /*
Default client encoding to use.
The initial value is set from the PGCLIENTENCODING environment variable.
*/ );
#endif
}
